# -------------------------------------------------------------------------------
# Copyright (c) 2024 Siemens
# All Rights Reserved.
# Authors: thomas.graf@siemens.com
#
# Licensed under the MIT license.
# SPDX-License-Identifier: MIT
# -------------------------------------------------------------------------------

from typing import Any, Dict, Optional

from .base import BaseMixin


class ModerationRequestMixin(BaseMixin):
    def get_all_moderation_requests(self, page: int = -1, page_size: int = -1,
                                    sort: str = "") -> Optional[Dict[str, Any]]:
        """Get information of about all moderation requests

        API endpoint: GET /moderationrequest

        :param page: page to retrieve
        :type page: int
        :param page_size: page size to use
        :type page_size: int
        :param sort: sort order for the packages ("name,desc"; "name,asc")
        :type sort: str
        :return: list of moderation requests
        :rtype: list of JSON moderation requests objects
        :raises SW360Error: if there is a negative HTTP response
        """

        full_url = self.url + "resource/api/moderationrequest"
        if page > -1:
            full_url = self._add_param(full_url, "page=" + str(page))
            full_url = self._add_param(full_url, "page_entries=" + str(page_size))

        if sort:
            # ensure HTML encoding
            sort = sort.replace(",", "%2C")
            full_url = self._add_param(full_url, "sort=" + sort)

        resp = self.api_get(full_url)
        return resp

    def get_moderation_requests_by_state(self, state: str, all_details: bool = False,
                                         page: int = -1, page_size: int = -1,
                                         sort: str = "") -> Optional[Dict[str, Any]]:
        """Get information of about all moderation requests

        API endpoint: GET /moderationrequest/byState

        :param all_details: retrieve all package details (optional))
        :type all_details: bool
        :param page: page to retrieve
        :type page: int
        :param page_size: page size to use
        :type page_size: int
        :param sort: sort order for the packages ("name,desc"; "name,asc")
        :type sort: str
        :return: list of moderation requests
        :rtype: list of JSON moderation requests objects
        :raises SW360Error: if there is a negative HTTP response
        """

        full_url = self.url + "resource/api/moderationrequest/byState?state=" + state
        if all_details:
            full_url = self._add_param(full_url, "allDetails=true")

        if page > -1:
            full_url = self._add_param(full_url, "page=" + str(page))
            full_url = self._add_param(full_url, "page_entries=" + str(page_size))

        if sort:
            # ensure HTML encoding
            sort = sort.replace(",", "%2C")
            full_url = self._add_param(full_url, "sort=" + sort)

        resp = self.api_get(full_url)
        return resp

    def get_moderation_request(self, mr_id: str) -> Optional[Dict[str, Any]]:
        """Get information of about a moderation request

        API endpoint: GET /moderationrequest/{id}

        :param mr_id: the id of the moderation request to be requested
        :type mr_id: string
        :return: a moderation request
        :rtype: JSON moderation request object
        :raises SW360Error: if there is a negative HTTP response
        """

        resp = self.api_get(self.url + "resource/api/moderationrequest/" + mr_id)
        return resp



# -------------------------------------------------------------------------------
# Copyright (c) 2019-2022 Siemens
# Copyright (c) 2022 BMW CarIT GmbH
# All Rights Reserved.
# Authors: thomas.graf@siemens.com, gernot.hillier@siemens.com
# Authors: helio.chissini-de-castro@bmw.de
#
# Licensed under the MIT license.
# SPDX-License-Identifier: MIT
# -------------------------------------------------------------------------------

from typing import Any, Dict, List, Optional

import requests

from .base import BaseMixin
from .sw360error import SW360Error


class LicenseMixin(BaseMixin):
    def create_new_license(
        self,
        shortName: str,
        fullName: str,
        text: str,
        checked: bool = False,
        license_details: Dict[str, Any] = {},
    ) -> Any:
        """Create a new license

        API endpoint: POST /licenses

        :param shortName: short license name. i.e "MIT"
        :param fullName: descriptive license name
        :param text: license text
        :param checked: if license is checked
        :type shortname: string
        :type fullname: string
        :type text: string
        :type checked: bool
        :return: SW360 result
        :rtype: JSON SW360 result object
        :raises SW360Error: if there is a negative HTTP response
        """

        url = self.url + "resource/api/licenses"

        license_details["shortName"] = shortName
        license_details["fullName"] = fullName
        license_details["text"] = text
        license_details["checked"] = checked

        response = self.api_post(url, json=license_details)
        if response is not None:
            if response.ok:
                return response.json()
        raise SW360Error(response, url)

    def delete_license(self, license_shortname: str) -> Optional[bool]:
        """Delete an existing license

        API endpoint: PATCH /licenses

        :param license_shortname: license shortname as the id
        :type license_shortname: string
        :return: SW360 result
        :rtype: JSON SW360 result object
        :raises SW360Error: if there is a negative HTTP response
        """

        if not license_shortname:
            raise SW360Error(message="No license shortname provided!")

        url = self.url + "resource/api/licenses/" + license_shortname
        print(url)
        response = self.api_delete(url)
        if response is not None:
            if response.ok:
                return True
        return None

    def download_license_info(
        self, project_id: str, filename: str, generator: str = "XhtmlGenerator", variant: str = "DISCLOSURE"
    ) -> None:
        """Gets the license information, aka Readme_OSS for the project
        with the given id

        API endpoint: GET /projects

        :param project_id: the id of the project to be deleted
        :param filename: the filename to be used
        :type project_id: string
        :type filename: string
        """
        hdr = self.api_headers.copy()
        hdr["Accept"] = "application/*"
        url = (
            self.url
            + "resource/api/projects/"
            + project_id
            + "/licenseinfo?generatorClassName="
            + generator
            + "&variant="
            + variant
        )
        req = requests.get(url, allow_redirects=True, headers=hdr)
        open(filename, "wb").write(req.content)

    def get_all_licenses(self) -> List[Dict[str, Any]]:
        """Get information of about all licenses

        API endpoint: GET /licenses

        :return: list of licenses
        :rtype: list of JSON license objects
        :raises SW360Error: if there is a negative HTTP response
        """

        resp = self.api_get(self.url + "resource/api/licenses")
        if resp and "_embedded" in resp and "sw360:licenses" in resp["_embedded"]:
            return resp["_embedded"]["sw360:licenses"]

        return []

    def get_license(self, license_id: str) -> Optional[Dict[str, Any]]:
        """Get information of about a license

        API endpoint: GET /licenses/{id}

        :param license_id: the id of the license to be requested
        :type license_id: string
        :return: a license
        :rtype: JSON license object
        :raises SW360Error: if there is a negative HTTP response
        """

        resp = self.api_get(self.url + "resource/api/licenses/" + license_id)
        return resp



# -------------------------------------------------------------------------------
# Copyright (c) 2022-2023 BMW CarIT GmbH
# All Rights Reserved.
# Authors: helio.chissini-de-castro@bmw.de
#
# Licensed under the MIT license.
# SPDX-License-Identifier: MIT
# -------------------------------------------------------------------------------

from typing import Any, Dict, Optional, Union
from urllib.parse import urljoin

import requests
import urllib3
from requests.auth import HTTPBasicAuth
from requests.structures import CaseInsensitiveDict

from .sw360error import SW360Error


class SW360OAuth2:
    """SW360 OAuth2 Credentials
    Restore or create Sw360 oauth2 tokens from user/password if auth server is alive
    :param url: URL of the SW360 instance
    :param user: SW360 username
    :param password: SW360 password
    :type url: string
    :type user: string
    :type password: string
    """

    _client_id: str
    _client_secret: str
    _user: str
    _password: str
    _token: str
    _refresh_token: str
    _url: str

    def __init__(self, url: str, user: str, password: str) -> None:
        self._url, self._user, self._password = url, user, password
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        self.__get_credentials()

    def __get_credentials(self) -> None:
        """Return last valid credentials id and secret

        Raises:
            SW360Error: If is unable to authorize
        """
        data: Dict[str, Any]
        url = urljoin(self._url, "/authorization/client-management")
        auth = HTTPBasicAuth(self._user, self._password)

        try:
            response = requests.get(url, verify=False, auth=auth)
        except Exception as ex:
            raise SW360Error(None, url, message="Unable to connect to oauth2 service: " + repr(ex))

        data = response.json()[0]
        self._client_id = data["client_id"]
        self._client_secret = data["client_secret"]

    def create_client(self, description: str, writeable: bool = False) -> None:
        """Create an OAuth2 client

        Args:
            description (str): Some description of the client
            writeable (bool): Create the id read/writeable

        Raises:
            SW360Error: When unable to create a client
        """
        scope = ["READ"]
        if writeable:
            scope.append("WRITE")
        payload: Union[str, Any] = {
            "description": description,
            "authorities": ["BASIC"],
            "scope": scope,
            "access_token_validity": 3600,
            "refresh_token_validity": 3600,
        }
        auth = HTTPBasicAuth(self._user, self._password)
        headers = {"content-type": "application/json"}
        url = urljoin(self._url, "/authorization/client-management")

        try:
            requests.post(url, json=payload, verify=False, auth=auth, headers=headers)
        except Exception as ex:
            raise SW360Error(None, url, message="Can't create oauth client: " + repr(ex))

    def generate_token(self) -> str:  # type: ignore
        """Generate a new bearer token
        """
        self.__oauth_token(True)  # type: ignore

    def __token(self, create: bool = False) -> None:
        """Create or return Liferay Token
        :param create: If should create or restore active tokens
        :type create: bool
        """
        data: Dict[str, Any]
        url = urljoin(self._url, "/authorization/oauth/token")
        payload: Dict[str, Any] = {
            "grant_type": "password",
            "username": self._user,
            "password": self._password,
        }
        auth = HTTPBasicAuth(self._client_id, self._client_secret)

        headers: CaseInsensitiveDict[Any] = CaseInsensitiveDict()
        headers["Content-Type"] = "application/x-www-form-urlencoded"

        if create:
            response = requests.post(url, auth=auth, headers=headers, json=payload, verify=False)
        else:
            params = f"grant_type=password&username={self._user}&password={self._password}"
            response = requests.get(url, auth=auth, headers=headers, params=params, verify=False)

        data = response.json()
        self._token = data["access_token"]
        self._refresh_token = data["refresh_token"]

    @property
    def token(self) -> Optional[str]:
        """Return the valid oauth token"""
        self.__token()
        return self._token

    @property
    def refresh_token(self) -> str:
        """Return the valid oauth refresh token"""
        self.__token()
        return self._refresh_token

    @property
    def url(self) -> str:
        """Return current session object url"""
        return self._url



# -------------------------------------------------------------------------------
# Copyright (c) 2019-2023 Siemens
# Copyright (c) 2022 BMW CarIT GmbH
# All Rights Reserved.
# Authors: thomas.graf@siemens.com, gernot.hillier@siemens.com
# Authors: helio.chissini-de-castro@bmw.de
#
# Licensed under the MIT license.
# SPDX-License-Identifier: MIT
# -------------------------------------------------------------------------------

import json
from typing import Any, Dict, Optional

from requests import Response


class SW360Error(IOError):
    """Base exception for SW360 operations

    :param message: a general error message
    :param response: the response object returned by the requests call
    :param url: the URL where the error occurred
    :type message: string
    :type response: object
    :type url: string
    """

    def __init__(self, response: Optional[Response] = None, url: str = "", message: str = "") -> None:
        self.message: str = message
        self.response: Optional[Response] = response
        self.url: str = url
        self.details: Optional[Dict[str, Any]] = None

        try:
            if response is not None:
                self.details = json.loads(response.text)
        except json.JSONDecodeError:
            self.details = None

        if message:
            super().__init__(message)
        else:
            super().__init__(str(response))



# -------------------------------------------------------------------------------
# Copyright (c) 2019-2024 Siemens
# All Rights Reserved.
# Authors: thomas.graf@siemens.com, gernot.hillier@siemens.com
#
# Licensed under the MIT license.
# SPDX-License-Identifier: MIT
# -------------------------------------------------------------------------------

"""Python interface to the Siemens SW360 platform"""

from typing import Any, Dict, Optional

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from .attachments import AttachmentsMixin
from .clearing import ClearingMixin
from .components import ComponentsMixin
from .license import LicenseMixin
from .moderationrequests import ModerationRequestMixin
from .packages import PackagesMixin
from .project import ProjectMixin
from .releases import ReleasesMixin
from .sw360error import SW360Error
from .vendor import VendorMixin
from .vulnerabilities import VulnerabilitiesMixin

# Retry mechanism for rate limiting
adapter = HTTPAdapter(max_retries=Retry(
    total=5,
    status_forcelist=[429, 500, 502, 503, 504],
    respect_retry_after_header=True,
    backoff_factor=30,
    allowed_methods=["HEAD", "GET", "OPTIONS", "POST", "PUT", "PATCH"]
))
session_default = requests.Session()
session_default.mount("http://", adapter)
session_default.mount("https://", adapter)


class SW360(
    AttachmentsMixin,
    ClearingMixin,
    ComponentsMixin,
    LicenseMixin,
    ProjectMixin,
    ReleasesMixin,
    VendorMixin,
    VulnerabilitiesMixin,
    PackagesMixin,
    ModerationRequestMixin
):
    """Python interface to the Siemens SW360 platform

    Authentication against a running SW360 instance is performed using an API token.
    The token will be sent as HTTP header using the format
    `Authorization: <token_type> <token>`. Check your SW360 REST API
    documentation for details on needed type and how to get the token.
    token_type is "Bearer" for an OAuth workflow and "Token" for tokens
    generated via the SW360 UI.

    :param url: URL of the SW360 instance
    :param token: The SW360 REST API token (the cryptic string without
     "Authorization:" and `token_type`).
    :param oauth2: flag indicating whether this is an OAuth2 token
    :type url: string
    :type token: string
    :type oauth2: boolean
    """

    def __init__(
        self,
        url: str,
        token: str,
        oauth2: bool = False,
        session: Optional[requests.Session] = session_default
    ) -> None:
        """Constructor"""
        if url[-1] != "/":
            url += "/"
        self.url: str = url
        self.session: Optional[requests.Session] = session

        if oauth2:
            self.api_headers = {"Authorization": "Bearer " + token}
        else:
            self.api_headers = {"Authorization": "Token " + token}

        self.force_no_session = False

    def login_api(self, token: str = "") -> bool:
        """Login to SW360 REST API. This used to have a `token` parameter
        due to historic reasons which is ignored.

        You need to call this before any other method accessing SW360.

        :raises SW360Error: if the login fails
        """
        if not self.force_no_session:
            self.session.headers = self.api_headers.copy()  # type: ignore

        url = self.url + "resource/api/"
        try:
            if self.force_no_session:
                resp = requests.get(url, headers=self.api_headers)
            else:
                if self.session:
                    resp = self.session.get(url)

        except Exception as ex:
            raise SW360Error(None, url, message="Unable to login: " + repr(ex))

        if resp.ok:
            return True
        else:
            raise SW360Error(resp, url, message="Unable to login")

    def close_api(self) -> None:
        """A keep-alive HTTP session is used to access the SW360 REST API.
        This method allows to explicitly close the connection at a defined
        time. Normally, you don't need to call it - session is cleaned up
        automatically when needed."""
        if self.session:
            self.session.close()
            self.session = None

    def api_get_raw(self, url: str = "") -> str:
        """Request `url` from REST API and return raw result.

        :param url: the url to be requested
        :type url: string
        :return: the HTTP response
        :rtype: string
        :raises SW360Error: if there is a negative HTTP response
        """
        if (not self.force_no_session) and self.session is None:
            raise SW360Error(message="login_api needs to be called first")

        if self.force_no_session:
            response = requests.get(url, headers=self.api_headers)
        else:
            if self.session:
                response = self.session.get(url)

        if response.ok:
            return response.text

        raise SW360Error(response, url)

    # ----- Health -------------------------------------------------------

    def get_health_status(self) -> Optional[Dict[str, Any]]:
        """Get information about the service's health.

        API endpoint: GET /health

        :return: service health status
        :rtype: JSON health status object
        :raises SW360Error: if there is a negative HTTP response
        """
        resp = self.api_get(self.url + "resource/health/")
        return resp
